@@ -4,12 +4,13 @@
 from __future__ import unicode_literals
 
 '''Generate header file for nanopb from a ProtoBuf FileDescriptorSet.'''
-nanopb_version = "nanopb-0.4.2"
+nanopb_version = "nanopb-0.4.3"
 
 import sys
 import re
 import codecs
 import copy
+import itertools
 import tempfile
 import shutil
 import os
@@ -211,33 +212,38 @@ assert varint_max_size(128) == 2
 class EncodedSize:
     '''Class used to represent the encoded size of a field or a message.
     Consists of a combination of symbolic sizes and integer sizes.'''
-    def __init__(self, value = 0, symbols = [], declarations = []):
+    def __init__(self, value = 0, symbols = [], declarations = [], required_defines = []):
         if isinstance(value, EncodedSize):
             self.value = value.value
             self.symbols = value.symbols
             self.declarations = value.declarations
+            self.required_defines = value.required_defines
         elif isinstance(value, strtypes + (Names,)):
             self.symbols = [str(value)]
             self.value = 0
             self.declarations = []
+            self.required_defines = [str(value)]
         else:
             self.value = value
             self.symbols = symbols
             self.declarations = declarations
+            self.required_defines = required_defines
 
     def __add__(self, other):
         if isinstance(other, int):
-            return EncodedSize(self.value + other, self.symbols, self.declarations)
+            return EncodedSize(self.value + other, self.symbols, self.declarations, self.required_defines)
         elif isinstance(other, strtypes + (Names,)):
-            return EncodedSize(self.value, self.symbols + [str(other)], self.declarations)
+            return EncodedSize(self.value, self.symbols + [str(other)], self.declarations, self.required_defines + [str(other)])
         elif isinstance(other, EncodedSize):
-            return EncodedSize(self.value + other.value, self.symbols + other.symbols, self.declarations + other.declarations)
+            return EncodedSize(self.value + other.value, self.symbols + other.symbols,
+                               self.declarations + other.declarations, self.required_defines + other.required_defines)
         else:
             raise ValueError("Cannot add size: " + repr(other))
 
     def __mul__(self, other):
         if isinstance(other, int):
-            return EncodedSize(self.value * other, [str(other) + '*' + s for s in self.symbols])
+            return EncodedSize(self.value * other, [str(other) + '*' + s for s in self.symbols],
+                               self.declarations, self.required_defines)
         else:
             raise ValueError("Cannot multiply size: " + repr(other))
 
@@ -248,8 +254,18 @@ class EncodedSize:
             return '(' + str(self.value) + ' + ' + ' + '.join(self.symbols) + ')'
 
     def get_declarations(self):
+        '''Get any declarations that must appear alongside this encoded size definition,
+        such as helper union {} types.'''
         return '\n'.join(self.declarations)
 
+    def get_cpp_guard(self, local_defines):
+        '''Get an #if preprocessor statement listing all defines that are required for this definition.'''
+        needed = [x for x in self.required_defines if x not in local_defines]
+        if needed:
+            return '#if ' + ' && '.join(['defined(%s)' % x for x in needed]) + "\n"
+        else:
+            return ''
+
     def upperlimit(self):
         if not self.symbols:
             return self.value
@@ -977,7 +993,8 @@ class OneOf(Field):
             # submessages.
             union_name = "%s_%s_size_union" % (self.struct_name, self.name)
             union_def = 'union %s {%s};\n' % (union_name, ' '.join('char f%d[%s];' % (k, s) for k,s in dynamic_sizes.items()))
-            return EncodedSize(0, ['sizeof(%s)' % union_name], [union_def])
+            required_defs = list(itertools.chain.from_iterable(s.required_defines for k,s in dynamic_sizes.items()))
+            return EncodedSize(0, ['sizeof(%s)' % union_name], [union_def], required_defs)
 
     def has_callbacks(self):
         return bool([f for f in self.fields if f.has_callbacks()])
@@ -1426,28 +1443,26 @@ class ProtoFile:
 
 
         def create_name(names):
-            if mangle_names == nanopb_pb2.M_NONE or mangle_names == nanopb_pb2.M_PACKAGE_INITIALS:
+            if mangle_names in (nanopb_pb2.M_NONE, nanopb_pb2.M_PACKAGE_INITIALS):
                 return base_name + names
-            elif mangle_names == nanopb_pb2.M_STRIP_PACKAGE:
+            if mangle_names == nanopb_pb2.M_STRIP_PACKAGE:
                 return Names(names)
-            else:
-                single_name = names
-                if isinstance(names, Names):
-                    single_name = names.parts[-1]
-                return Names(single_name)
+            single_name = names
+            if isinstance(names, Names):
+                single_name = names.parts[-1]
+            return Names(single_name)
 
         def mangle_field_typename(typename):
             if mangle_names == nanopb_pb2.M_FLATTEN:
                 return "." + typename.split(".")[-1]
-            elif strip_prefix is not None and typename.startswith(strip_prefix):
+            if strip_prefix is not None and typename.startswith(strip_prefix):
                 if replacement_prefix is not None:
                     return "." + replacement_prefix + typename[len(strip_prefix):]
                 else:
                     return typename[len(strip_prefix):]
-            elif self.file_options.package:
+            if self.file_options.package:
                 return "." + replacement_prefix + typename
-            else:
-                return typename
+            return typename
 
         if replacement_prefix is not None:
             base_name = Names(replacement_prefix.split('.'))
@@ -1543,8 +1558,12 @@ class ProtoFile:
         yield '\n'
 
         for incfile in self.file_options.include:
-            yield options.genformat % incfile
-            yield '\n'
+            # allow including system headers
+            if (incfile.startswith('<')):
+                yield '#include %s\n' % incfile
+            else:
+                yield options.genformat % incfile
+                yield '\n'
 
         for incfile in includes:
             noext = os.path.splitext(incfile)[0]
@@ -1561,10 +1580,6 @@ class ProtoFile:
         yield '#endif\n'
         yield '\n'
 
-        yield '#ifdef __cplusplus\n'
-        yield 'extern "C" {\n'
-        yield '#endif\n\n'
-
         if self.enums:
             yield '/* Enum definitions */\n'
             for enum in self.enums:
@@ -1589,6 +1604,10 @@ class ProtoFile:
                     yield enum.auxiliary_defines() + '\n'
                 yield '\n'
 
+        yield '#ifdef __cplusplus\n'
+        yield 'extern "C" {\n'
+        yield '#endif\n\n'
+
         if self.messages:
             yield '/* Initializer values for message structs */\n'
             for msg in self.messages:
@@ -1620,12 +1639,21 @@ class ProtoFile:
             yield '\n'
 
             yield '/* Maximum encoded size of messages (where known) */\n'
+            messagesizes = []
             for msg in self.messages:
-                msize = msg.encoded_size(self.dependencies)
                 identifier = '%s_size' % msg.name
+                messagesizes.append((identifier, msg.encoded_size(self.dependencies)))
+
+            # If we require a symbol from another file, put a preprocessor if statement
+            # around it to prevent compilation errors if the symbol is not actually available.
+            local_defines = [identifier for identifier, msize in messagesizes if msize is not None]
+            for identifier, msize in messagesizes:
                 if msize is not None:
+                    cpp_guard = msize.get_cpp_guard(local_defines)
+                    yield cpp_guard
                     yield msize.get_declarations()
                     yield '#define %-40s %s\n' % (identifier, msize)
+                    if cpp_guard: yield "#endif\n"
                 else:
                     yield '/* %s depends on runtime parameters */\n' % identifier
             yield '\n'
@@ -2006,8 +2034,8 @@ def main_cli():
         else:
             data = open(filename, 'rb').read()
 
-        for fdesc in descriptor.FileDescriptorSet.FromString(data).file:
-            fdescs[fdesc.name] = fdesc
+        fdesc = descriptor.FileDescriptorSet.FromString(data).file[-1]
+        fdescs[fdesc.name] = fdesc
 
     # Process any include files first, in order to have them
     # available as dependencies
@@ -2031,7 +2059,10 @@ def main_cli():
             sys.stderr.write("Writing to %s\n" % paths)
 
         for path, data in to_write:
-            os.makedirs(os.path.dirname(path), exist_ok=True)
+            dirname = os.path.dirname(path)
+            if dirname and not os.path.exists(dirname):
+                os.makedirs(dirname)
+
             with open(path, 'w') as f:
                 f.write(data)
 
